from datetime import datetime
from typing import Dict, List, Tuple

from i8_terminal.common.metrics import get_all_metrics_types_dict
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice

PERIODS: Dict[str, List[Tuple[str, str]]] = {
    "summary": [],
    "price": [*[(f"{i}da", f"{i} day{'s' if i > 1 else ''} ago") for i in range(1, 3)]],
    "performance": [*[(f"{i}da", f"{i} day{'s' if i > 1 else ''} ago") for i in range(1, 3)]],
    "technical": [*[(f"{i}da", f"{i} day{'s' if i > 1 else ''} ago") for i in range(1, 3)]],
    "fin_metric": [
        ("mrq", "most recent quarter"),
        ("mry", "most recent financial year"),
        ("ttm", "trailing 12 months"),
        ("ytd", "year to date"),
        *[(f"{i}qa", f"{i} quarter{'s' if i > 1 else ''} ago") for i in range(1, 3)],
        *[(f"{i}ya", f"{i} year{'s' if i > 1 else ''} ago") for i in range(1, 3)],
        *[(f"q{i}_{y}", f"Fiscal quarter {i} {y}") for i in range(1, 5) for y in range(datetime.now().year, 2008, -1)],
        *[
            (f"q{i}ttm_{y}", f"Fiscal quarter {i} {y}")
            for i in range(1, 4)
            for y in range(datetime.now().year, 2008, -1)
        ],
        *[
            (f"q{i}ytd_{y}", f"Fiscal quarter {i} {y}")
            for i in range(2, 4)
            for y in range(datetime.now().year, 2008, -1)
        ],
        *[(f"fy_{y}", f"Fiscal year {y}") for y in range(datetime.now().year, 2008, -1)],
    ],
    "fin_statement": [
        ("mrq", "most recent quarter"),
        ("mry", "most recent financial year"),
        ("ttm", "trailing 12 months"),
        ("ytd", "year to date"),
        *[(f"{i}qa", f"{i} quarter{'s' if i > 1 else ''} ago") for i in range(1, 3)],
        *[(f"{i}ya", f"{i} year{'s' if i > 1 else ''} ago") for i in range(1, 3)],
        *[(f"q{i}_{y}", f"Fiscal quarter {i} {y}") for i in range(1, 5) for y in range(datetime.now().year, 2008, -1)],
        *[
            (f"q{i}ttm_{y}", f"Fiscal quarter {i} {y}")
            for i in range(1, 4)
            for y in range(datetime.now().year, 2008, -1)
        ],
        *[
            (f"q{i}ytd_{y}", f"Fiscal quarter {i} {y}")
            for i in range(2, 4)
            for y in range(datetime.now().year, 2008, -1)
        ],
        *[(f"fy_{y}", f"Fiscal year {y}") for y in range(datetime.now().year, 2008, -1)],
    ],
    "earnings": [
        ("mrq", "most recent quarter"),
        *[(f"{i}qa", f"{i} quarter{'s' if i > 1 else ''} ago") for i in range(1, 3)],
        *[(f"q{i}_{y}", f"Fiscal quarter {i} {y}") for i in range(1, 5) for y in range(datetime.now().year, 2008, -1)],
        ("uq", "upcoming quarter"),
    ],
}


class PeriodParamType(AutoCompleteChoice):
    name = "period"
    metrics: Dict[str, str] = {}

    def get_suggestions(self, keyword: str, pre_populate: bool = True, metric: str = None) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.metrics = get_all_metrics_types_dict()
        self.set_choices(PERIODS.get(self.metrics[metric if metric else ""], []))

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "PERIOD"



from typing import Optional

import click
import investor8_sdk
from pandas import DataFrame
from rich.console import Console

from i8_terminal.commands.earnings import earnings
from i8_terminal.common.cli import pass_command
from i8_terminal.common.formatting import get_formatter
from i8_terminal.common.layout import df2Table, format_df
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.common.utils import export_data, export_to_html
from i8_terminal.config import APP_SETTINGS
from i8_terminal.types.ticker_param_type import TickerParamType


def get_upcoming_earnings_df(size: int) -> DataFrame:
    earnings = investor8_sdk.EarningsApi().get_upcoming_earnings(size=size)
    earnings = [d.to_dict() for d in earnings]
    df = DataFrame(earnings)
    df["eps_beat_rate"] = df["eps_beat_rate"] * 100
    df["revenue_beat_rate"] = df["revenue_beat_rate"] * 100
    return df


def get_upcoming_earnings_df_by_ticker(tickers: str) -> DataFrame:
    upcoming_earnings = []
    for tk in tickers.replace(" ", "").upper().split(","):
        upcoming_earnings.extend([investor8_sdk.EarningsApi().get_upcoming_earning(tk)])
    df = DataFrame([h.to_dict() for h in upcoming_earnings])
    df["eps_beat_rate"] = df["eps_beat_rate"] * 100
    df["revenue_beat_rate"] = df["revenue_beat_rate"] * 100
    return df


def format_upcoming_earnings_df(df: DataFrame, target: str) -> DataFrame:
    formatters = {
        "latest_price": get_formatter("price", target),
        "change": get_formatter("perc", target),
        "fyq": get_formatter("fyq", target),
        "eps_ws": get_formatter("number", target),
        "eps_beat_rate": get_formatter("number_perc" if target == "console" else "perc", target),
        "revenue_ws": get_formatter("financial", target),
        "revenue_beat_rate": get_formatter("number_perc" if target == "console" else "perc", target),
    }
    col_names = {
        "ticker": "Ticker",
        "name": "Name",
        "latest_price": "Price",
        "change": "Change",
        "actual_report_date": "Report Date",
        "fyq": "Period",
        "call_time": "Call Time",
        "eps_ws": "EPS Estimate",
        "eps_beat_rate": "EPS Beat Rate",
        "revenue_ws": "Revenue Estimate",
        "revenue_beat_rate": "Revenue Beat Rate",
    }
    return format_df(df, col_names, formatters)


def format_upcoming_earnings_df_by_ticker(df: DataFrame, target: str) -> DataFrame:
    formatters = {
        "fyq": get_formatter("fyq", target),
        "eps_ws": get_formatter("number", target),
        "eps_beat_rate": get_formatter("number_perc" if target == "console" else "perc", target),
        "revenue_ws": get_formatter("financial", target),
        "revenue_beat_rate": get_formatter("number_perc" if target == "console" else "perc", target),
    }
    col_names = {
        "ticker": "Ticker",
        "actual_report_date": "Report Date",
        "fyq": "Period",
        "call_time": "Call Time",
        "eps_ws": "EPS Estimate",
        "eps_beat_rate": "EPS Beat Rate",
        "revenue_ws": "Revenue Estimate",
        "revenue_beat_rate": "Revenue Beat Rate",
    }
    return format_df(df, col_names, formatters)


@earnings.command()
@click.option(
    "--tickers", "-k", type=TickerParamType(), callback=validate_tickers, help="Comma-separated list of tickers."
)
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@pass_command
def upcoming(tickers: Optional[str], export_path: Optional[str]) -> None:
    """
    Lists upcoming company earnings.

    Examples:

    `i8 earnings upcoming`

    `i8 earnings upcoming --tickers AAPL,MSFT`

    """
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        df = get_upcoming_earnings_df_by_ticker(tickers) if tickers else get_upcoming_earnings_df(size=20)
    if export_path:
        if export_path.split(".")[-1] == "html":
            df_formatted = (
                format_upcoming_earnings_df_by_ticker(df, "console")
                if tickers
                else format_upcoming_earnings_df(df, "console")
            )
            table = df2Table(df_formatted)
            export_to_html(table, export_path)
            return
        export_data(
            format_upcoming_earnings_df_by_ticker(df, "store") if tickers else format_upcoming_earnings_df(df, "store"),
            export_path,
            column_width=18,
            column_format=APP_SETTINGS["styles"]["xlsx"]["financials"]["column"],
        )
        return
    df_formatted = (
        format_upcoming_earnings_df_by_ticker(df, "console") if tickers else format_upcoming_earnings_df(df, "console")
    )
    table = df2Table(df_formatted)
    console.print(table)



from typing import List, Tuple

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_statements() -> List[Tuple[str, str]]:
    return [
        ("income", "Income Statment"),
        ("cash_flow", "Cash Flow Statment"),
        ("balance_sheet", "Balance Sheet Statment"),
    ]


class FinancialStatementParamType(AutoCompleteChoice):
    name = "financialstatement"

    def get_suggestions(self, keyword: str, pre_populate: bool = True) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_statements())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "FINANCIALSTATEMENT"



import investor8_sdk
from pandas import DataFrame

from i8_terminal.common.utils import status
from i8_terminal.service_result.column_info import ColumnInfo
from i8_terminal.service_result.columns_context import ColumnsContext
from i8_terminal.service_result.earning_list_result import EarningsListResult


@status()
def get_earnings_list(ticker: str, size: int) -> EarningsListResult:
    historical_earnings = investor8_sdk.EarningsApi().get_historical_earnings(ticker, size=size)
    historical_earnings = [d.to_dict() for d in historical_earnings]
    df = DataFrame(historical_earnings)
    df["period"] = df.fyq.str[2:-2] + " " + df.fyq.str[-2:]
    df.rename(
        columns={
            "actual_report_time": "earning_date",
            "eps_ws": "eps_consensus",
            "revenue_ws": "revenue_consensus",
            "call_time": "earning_call_time",
        },
        inplace=True,
    )
    cc = ColumnsContext(
        [
            ColumnInfo(name="period", col_type="context", display_name="Period", data_type="str", unit="string"),
            ColumnInfo(name="earning_date", col_type="metric"),
            ColumnInfo(name="earning_call_time", col_type="metric"),
            ColumnInfo(name="eps_consensus", col_type="metric"),
            ColumnInfo(name="eps_actual", col_type="metric"),
            ColumnInfo(name="eps_surprise", col_type="metric"),
            ColumnInfo(name="revenue_consensus", col_type="metric"),
            ColumnInfo(name="revenue_actual", col_type="metric"),
            ColumnInfo(name="revenue_surprise", col_type="metric"),
        ]
    )
    return EarningsListResult(df, cc)



from pydoc import locate
from typing import Any, Dict

import numpy as np
from pandas import DataFrame
from rich.table import Table

from i8_terminal.common.formatting import get_formatter
from i8_terminal.config import get_table_style


def format_df(df: DataFrame, cols_map: Dict[str, str], cols_formatters: Dict[str, Any]) -> DataFrame:
    for c, f in cols_formatters.items():
        df[c] = df[c].map(f)
    return df[cols_map.keys()].rename(columns=cols_map)


def format_metrics_df(df: DataFrame, target: str) -> DataFrame:
    df["value"] = df.apply(
        lambda metric: get_formatter(
            "number_int"
            if metric.data_format == "int" and metric.display_format == "number"
            else metric.display_format,
            target,
        )(
            locate(metric.data_format)(locate("float")(metric.value) if metric.data_format == "int" else metric.value)  # type: ignore # noqa: E501
        ),
        axis=1,
    )
    return df


def df2Table(df: DataFrame, style_profile: str = "default", columns_justify: Dict[str, Any] = {}) -> Table:
    MIN_COL_LENGTH = 13
    style = get_table_style(style_profile)
    table = Table(**style)
    default_justify = {
        "Price": "right",
        "Open": "right",
        "Close": "right",
        "Low": "right",
        "High": "right",
        "Volume": "right",
        "Change": "right",
        "Change (%)": "right",
        "Market Cap": "right",
        "EPS Cons.": "right",
        "EPS Actual": "right",
        "Revenue Cons.": "right",
        "Revenue Actual": "right",
        "Level": "right",
        "EPS Estimate": "right",
        "Revenue Estimate": "right",
        "EPS Beat Rate": "right",
        "Revenue Beat Rate": "right",
        "EPS Surprise": "right",
        "Revenue Surprise": "right",
        "EPS Consensus": "right",
        "Revenue Consensus": "right",
        "Eps Surprise": "right",
    }
    for c in df.columns:
        table.add_column(
            c,
            justify=columns_justify.get(c, default_justify.get(c, "left")),
            min_width=min(max(df[c].str.len().max(), len(df[c].name)), MIN_COL_LENGTH),
        )
    for _, r in df.iterrows():
        row = [r[c] if r[c] is not np.nan and r[c] is not None else "-" for c in df.columns]
        table.add_row(*row)
    return table



import enum
import os
from difflib import SequenceMatcher
from io import StringIO
from typing import Any, Callable, Dict, Optional, TypeVar

import arrow
import pandas as pd
from rich.console import Console

from i8_terminal.config import APP_SETTINGS

T = TypeVar("T")


class PlotType(enum.Enum):
    CHART = "chart"
    TABLE = "table"


def to_snake_case(value: str) -> str:
    return "_".join(value.lower().split())


def get_period_code(period: str) -> int:
    return {"1D": 1, "5D": 2, "1M": 3, "3M": 4, "6M": 5, "1Y": 6, "3Y": 7, "5Y": 8}.get(period, 3)


def get_period_days(period: str) -> int:
    return {"1D": 1, "5D": 5, "1M": 30, "3M": 90, "6M": 180, "1Y": 365, "3Y": 1095, "5Y": 1825}.get(period, 365)


def similarity(a: str, b: str) -> float:
    return SequenceMatcher(None, a, b).ratio()


def export_data(
    export_df: pd.DataFrame,
    export_path: str,
    column_width: Optional[int],
    column_format: Dict[str, Any],
    index: bool = False,
) -> None:
    console = Console()
    extension = export_path.split(".")[-1]
    if extension == "csv":
        export_df.to_csv(export_path, index=index)
        console.print(f"Data is saved on: {export_path}")
    elif extension == "xlsx":
        writer = pd.ExcelWriter(export_path, engine="xlsxwriter")
        export_df.to_excel(writer, sheet_name="Sheet1", startrow=1, header=False, index=index)
        workbook = writer.book
        worksheet = writer.sheets["Sheet1"]
        header_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["header"])
        metric_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["metric"])
        column_format = workbook.add_format(column_format)
        headers = export_df.columns.tolist()
        if index:
            headers.insert(0, (export_df.index.name, ""))
        for col_num, value in enumerate(headers):
            if type(value) is tuple:
                worksheet.merge_range(0, col_num, len(value) - 1, col_num, " ".join(reversed(value)), header_format)
            else:
                worksheet.write(0, col_num, value, header_format)
        worksheet.set_column(0, 0, 20, metric_format)
        worksheet.set_column(
            1, len(export_df.columns) - 1 if not index else len(export_df.columns), column_width, column_format
        )
        writer.save()
        console.print(f"Data is saved on: {export_path}")
    else:
        console.print("export_path is not valid")


def is_cached_file_expired(file_path: str) -> bool:
    mtime = arrow.get(os.path.getmtime(file_path))
    return bool(mtime < arrow.utcnow().shift(hours=-APP_SETTINGS.get("cache", {}).get("age", 48)))


def reverse_period(period: str) -> str:
    """
    If period is fyq type (eg. 'Q 2021'), the function will change it to '2021 Q'.
    """
    splitted_period = period.split(" ")
    return f"{splitted_period[1]} {splitted_period[0]}" if len(splitted_period) > 1 else period


def export_to_html(data: Any, export_path: str) -> None:
    console = Console(record=True, file=StringIO())
    console.print(data)
    exported_html = console.export_html(
        inline_styles=True,
        code_format="""
        <html>
            <head>
                <title>i8 Terminal by Investor8</title>
            </head>
            <body>
                <div>
                    <img src='https://i8terminal.io/img/TerminalLogo.png' width='16%'>
                </div>
                <div>
                    <pre>{code}</pre>
                </div
            </body>
        </html>
        """,
    )
    with open(export_path, "w", encoding="utf-8") as file:
        file.write(exported_html)
    console = Console()
    console.print(f"Data is saved on: {export_path}")


def status(text: str = "Fetching data...", spinner: str = "material") -> Callable[..., Callable[..., T]]:
    def decorate(func: Any) -> Any:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            console = Console()
            with console.status(text, spinner=spinner):
                return func(*args, **kwargs)

        return wrapper

    return decorate
